#
# Server support for DBus properties
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
# For more info about DBus specification see:
# https://dbus.freedesktop.org/doc/dbus-specification.html#introspection-format
#
from abc import ABCMeta
from collections import defaultdict
from functools import wraps
from typing import Dict

from dasbus.server.interface import dbus_signal, get_xml
from dasbus.specification import DBusSpecification, DBusSpecificationError
from dasbus.typing import get_variant, Str, Variant, List

__all__ = [
    "emits_properties_changed",
    "PropertiesException",
    "PropertiesInterface"
]


def emits_properties_changed(method):
    """Decorator for emitting properties changes.

    The decorated method has to be a member of a class that
    inherits PropertiesInterface.

    :param method: a DBus method of a class that inherits PropertiesInterface
    :return: a wrapper of a DBus method that emits PropertiesChanged
    """
    @wraps(method)
    def wrapper(obj, *args, **kwargs):
        result = method(obj, *args, **kwargs)
        obj.flush_changes()
        return result

    return wrapper


class PropertiesException(Exception):
    """Exception for DBus properties."""
    pass


class PropertiesChanges(object):
    """Cache for properties changes.

    This class is useful to collect the changed properties
    and their values, before they are emitted on DBus.
    """

    def __init__(self, obj):
        """Create the cache.

        :param obj: an object with DBus properties
        """
        self._object = obj
        self._properties_names = set()
        self._properties_specs = self._find_properties_specs(obj)

    def _find_properties_specs(self, obj):
        """Find specifications of DBus properties.

        :param obj: an object with DBus properties
        :return: a map of property names and their specifications
        """
        specification = DBusSpecification.from_xml(get_xml(obj))
        properties_specs = {}

        for member in specification.members:
            if not isinstance(member, DBusSpecification.Property):
                continue

            if member.name in properties_specs:
                raise DBusSpecificationError(
                    "DBus property '{}' is defined in more than "
                    "one interface.".format(member.name)
                )

            properties_specs[member.name] = member

        return properties_specs

    def flush(self):
        """Flush the cache.

        The content of the cache will be composed to requests
        and the cache will be cleared.

        The requests can be used to emit the PropertiesChanged
        signal. The requests are a list of tuples, that contain
        an interface name and a dictionary of properties changes.

        :return: a list of requests
        """
        content = self._properties_names
        self._properties_names = set()
        requests = defaultdict(dict)

        for property_name in content:
            # Find the property specification.
            member = self._properties_specs[property_name]

            # Get the property value.
            value = getattr(self._object, property_name)
            variant = get_variant(member.type, value)

            # Create a request.
            requests[member.interface_name][member.name] = variant

        return requests.items()

    def check_property(self, property_name):
        """Check if the property name is valid."""
        if property_name not in self._properties_specs:
            raise PropertiesException(
                "DBus object has no property '{}'.".format(
                    property_name
                )
            )

    def update(self, property_name):
        """Update the cache."""
        self.check_property(property_name)
        self._properties_names.add(property_name)


class PropertiesInterface(metaclass=ABCMeta):
    """Standard DBus interface org.freedesktop.DBus.Properties.

    DBus objects don't have to inherit this class, because the
    DBus library provides support for this interface by default.
    This class only extends this support.

    Report the changed property:

    .. code-block:: python

        self.report_changed_property('X')

    Emit all changes when the method is done:

    .. code-block:: python

        @emits_properties_changed
        def SetX(x: Int):
            self.set_x(x)

    """

    def __init__(self):
        """Initialize the interface."""
        self._properties_changes = PropertiesChanges(self)

    @dbus_signal
    def PropertiesChanged(self, interface: Str, changed: Dict[Str, Variant],
                          invalid: List[Str]):
        """Standard signal properties changed.

        :param interface: a name of an interface
        :param changed: a dictionary of changed properties
        :param invalid: a list of invalidated properties
        :return:
        """
        pass

    def report_changed_property(self, property_name):
        """Reports changed DBus property.

        :param property_name: a name of a DBus property
        """
        self._properties_changes.update(property_name)

    def flush_changes(self):
        """Flush properties changes."""
        requests = self._properties_changes.flush()

        for interface, changes in requests:
            self.PropertiesChanged(interface, changes, [])



#
# Support for DBus errors
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from abc import ABCMeta, abstractmethod

from dasbus.namespace import get_dbus_name

__all__ = [
    "get_error_decorator",
    "DBusError",
    "AbstractErrorRule",
    "ErrorRule",
    "DefaultErrorRule",
    "ErrorMapper"
]


def get_error_decorator(error_mapper):
    """Generate a decorator for DBus errors.

    Create a function for decorating Python exception classes.
    The decorator will add a new rule to the given error mapper
    that will map the class to the specified error name.

    Definition of the decorator:

    .. code-block:: python

        decorator(error_name, namespace=())

    The decorator accepts a name of the DBus error and optionally
    a namespace of the DBus name. The namespace will be used as
    a prefix of the DBus name.

    Usage:

    .. code-block:: python

        # Create an error mapper.
        error_mapper = ErrorMapper()

        # Create a decorator for DBus errors and use it to map
        # the class ExampleError to the name my.example.Error.
        dbus_error = create_error_decorator(error_mapper)

        @dbus_error("my.example.Error")
        class ExampleError(DBusError):
            pass

    :param error_mapper: an error mapper
    :return: a decorator
    """
    def decorator(error_name, namespace=()):
        error_name = get_dbus_name(*namespace, error_name)

        def decorated(cls):
            error_mapper.add_rule(ErrorRule(
                exception_type=cls,
                error_name=error_name
            ))
            return cls

        return decorated

    return decorator


class DBusError(Exception):
    """A default DBus error."""
    pass


class AbstractErrorRule(metaclass=ABCMeta):
    """Abstract rule for mapping a Python exception to a DBus error."""

    __slots__ = []

    @abstractmethod
    def match_type(self, exception_type):
        """Is this rule matching the given exception type?

        :param exception_type: a type of the Python error
        :return: True or False
        """
        pass

    @abstractmethod
    def get_name(self, exception_type):
        """Get a DBus name for the given exception type.

        :param exception_type: a type of the Python error
        :return: a name of the DBus error
        """
        pass

    @abstractmethod
    def match_name(self, error_name):
        """Is this rule matching the given DBus error?

        :param error_name: a name of the DBus error
        :return: True or False
        """
        pass

    @abstractmethod
    def get_type(self, error_name):
        """Get an exception type of the given DBus error.

        param error_name: a name of the DBus error
        :return: a type of the Python error
        """
        pass


class ErrorRule(AbstractErrorRule):
    """Rule for mapping a Python exception to a DBus error."""

    __slots__ = [
        "_exception_type",
        "_error_name"
    ]

    def __init__(self, exception_type, error_name):
        """Create a new error rule.

        The rule will return the Python type exception_type
        for the DBue error error_name.

        The rule will return the DBue name error_name for
        the Python type exception_type

        :param exception_type: a type of the Python error
        :param error_name: a name of the DBus error
        """
        self._exception_type = exception_type
        self._error_name = error_name

    def match_type(self, exception_type):
        """Is this rule matching the given exception type?"""
        return self._exception_type == exception_type

    def get_name(self, exception_type):
        """Get a DBus name for the given exception type."""
        return self._error_name

    def match_name(self, error_name):
        """Is this rule matching the given DBus error?"""
        return self._error_name == error_name

    def get_type(self, error_name):
        """Get an exception type of the given DBus error."""
        return self._exception_type


class DefaultErrorRule(AbstractErrorRule):
    """Default rule for mapping a Python exception to a DBus error."""

    __slots__ = [
        "_default_type",
        "_default_namespace"
    ]

    def __init__(self, default_type, default_namespace):
        """Create a new default rule.

        The rule will return the Python type default_type
        for the all DBus errors.

        The rule will generate a DBus name with the prefix
        default_namespace for all Python exception types.

        :param default_type: a default type of the Python error
        :param default_namespace: a default namespace of the DBus error
        """
        self._default_type = default_type
        self._default_namespace = default_namespace

    def match_type(self, exception_type):
        """Is this rule matching the given exception type?"""
        return True

    def get_name(self, exception_type):
        """Get a DBus name for the given exception type."""
        return get_dbus_name(*self._default_namespace, exception_type.__name__)

    def match_name(self, error_name):
        """Is this rule matching the given DBus error?"""
        return True

    def get_type(self, error_name):
        """Get an exception type of the given DBus error."""
        return self._default_type


class ErrorMapper(object):
    """Class for mapping Python exceptions to DBus errors."""

    __slots__ = ["_error_rules"]

    def __init__(self):
        """Create a new error mapper."""
        self._error_rules = []
        self.reset_rules()

    def add_rule(self, rule: AbstractErrorRule):
        """Add a rule to the error mapper.

        The new rule will have a higher priority than
        the rules already contained in the error mapper.

        :param rule: an error rule
        :type rule: an instance of AbstractErrorRule
        """
        self._error_rules.append(rule)

    def reset_rules(self):
        """Reset rules in the error mapper.

        Reset the error rules to the initial state.
        All rules will be replaced with the default ones.
        """
        # Clear the list.
        self._error_rules = []

        # Add the default rules.
        self.add_rule(DefaultErrorRule(
            default_type=DBusError,
            default_namespace=("not", "known", "Error")
        ))

    def get_error_name(self, exception_type):
        """Get a DBus name of the Python exception.

        Try to find a matching rule in the error mapper.
        If a rule matches the given exception type, use
        the rule to get the name of the DBus error.

        The rules in the error mapper are processed in
        the reversed order to respect the priority of
        the rules.

        :param exception_type: a type of the Python error
        :type exception_type: a subclass of Exception
        :return: a name of the DBus error
        :raise LookupError: if no name is found
        """
        for rule in reversed(self._error_rules):
            if rule.match_type(exception_type):
                return rule.get_name(exception_type)

        raise LookupError(
            "No name found for '{}'.".format(exception_type.__name__)
        )

    def get_exception_type(self, error_name):
        """Get a Python exception type of the DBus error.

        Try to find a matching rule in the error mapper.
        If a rule matches the given name of a DBus error,
        use the rule to get the type of a Python exception.

        The rules in the error mapper are processed in
        the reversed order to respect the priority of
        the rules.

        :param error_name: a name of the DBus error
        :return: a type of the Python exception
        :rtype: a subclass of Exception
        :raise LookupError: if no type is found
        """
        for rule in reversed(self._error_rules):
            if rule.match_name(error_name):
                return rule.get_type(error_name)

        raise LookupError("No type found for '{}'.".format(error_name))



#
# Support for XML representation
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
# For more info about DBus specification see:
# https://dbus.freedesktop.org/doc/dbus-specification.html#introspection-format
#
from xml.etree import ElementTree
from xml.dom import minidom

__all__ = [
    "XMLParser",
    "XMLGenerator"
]


class XMLParser(object):
    """Class for parsing XML."""

    @staticmethod
    def xml_to_element(xml):
        return ElementTree.fromstring(xml)

    @staticmethod
    def is_member(member_node):
        return member_node.tag in ("method", "signal", "property")

    @staticmethod
    def is_interface(member_node):
        return member_node.tag == "interface"

    @staticmethod
    def is_signal(member_node):
        return member_node.tag == "signal"

    @staticmethod
    def is_method(member_node):
        return member_node.tag == "method"

    @staticmethod
    def is_property(member_node):
        return member_node.tag == "property"

    @staticmethod
    def is_parameter(member_node):
        return member_node.tag == "arg"

    @staticmethod
    def has_name(node, node_name):
        return node.attrib.get("name", "") == node_name

    @staticmethod
    def get_name(node):
        return node.attrib["name"]

    @staticmethod
    def get_type(node):
        return node.attrib["type"]

    @staticmethod
    def get_access(node):
        return node.attrib["access"]

    @staticmethod
    def get_direction(node):
        return node.attrib["direction"]

    @staticmethod
    def get_interfaces_from_node(node_element):
        """Return a dictionary of interfaces defined in a node element."""
        interfaces = {}

        for element in node_element.iterfind("interface"):
            interfaces[element.attrib["name"]] = element

        return interfaces


class XMLGenerator(XMLParser):
    """Class for generating XML."""

    @staticmethod
    def element_to_xml(element):
        """Return XML of the element."""
        return ElementTree.tostring(
            element,
            method="xml",
            encoding="unicode"
        )

    @staticmethod
    def prettify_xml(xml):
        """Return pretty printed normalized XML.

        Python 3.8 changed the order of the attributes and introduced
        the function canonicalize that should be used to normalize XML.
        """
        # Remove newlines and extra whitespaces,
        xml_line = "".join([line.strip() for line in xml.splitlines()])

        # Generate pretty xml.
        xml = minidom.parseString(xml_line).toprettyxml(indent="  ")

        # Normalize attributes.
        canonicalize = getattr(
            ElementTree, "canonicalize", lambda xml, *args, **kwargs: xml
        )

        return canonicalize(xml, with_comments=True)

    @staticmethod
    def add_child(parent_element, child_element):
        """Append the child element to the parent element."""
        parent_element.append(child_element)

    @staticmethod
    def add_comment(element, comment):
        element.append(ElementTree.Comment(text=comment))

    @staticmethod
    def create_node():
        """Create a node element called node."""
        return ElementTree.Element("node")

    @staticmethod
    def create_interface(name):
        """Create an interface element."""
        return ElementTree.Element("interface", {"name": name})

    @staticmethod
    def create_signal(name):
        """Create a signal element."""
        return ElementTree.Element("signal", {"name": name})

    @staticmethod
    def create_method(name):
        """Create a method element."""
        return ElementTree.Element("method", {"name": name})

    @staticmethod
    def create_parameter(name, param_type, direction):
        """Create a parameter element."""
        tag = "arg"
        attr = {
            "name": name,
            "type": param_type,
            "direction": direction
        }
        return ElementTree.Element(tag, attr)

    @staticmethod
    def create_property(name, property_type, access):
        """Create a property element."""
        tag = "property"
        attr = {
            "name": name,
            "type": property_type,
            "access": access
        }
        return ElementTree.Element(tag, attr)



#
# DBus names and paths
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA#

__all__ = [
    "get_dbus_name",
    "get_dbus_path",
    "get_namespace_from_name"
]


def get_dbus_name(*namespace):
    """Create a DBus name from the given names.

    :param namespace: a sequence of names
    :return: a DBus name
    """
    return ".".join(namespace)


def get_dbus_path(*namespace):
    """Create a DBus path from the given names.

    :param namespace: a sequence of names
    :return: a DBus path
    """
    return "/" + "/".join(namespace)


def get_namespace_from_name(name):
    """Return a namespace of the DBus name.

    :param name: a DBus name
    :return: a sequence of names
    """
    return tuple(name.split("."))



#
# Representation of DBus connections
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import logging
import threading
from abc import ABCMeta, abstractmethod

from dasbus.constants import DBUS_NAME_FLAG_ALLOW_REPLACEMENT, \
    DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER
from dasbus.client.proxy import ObjectProxy, InterfaceProxy
from dasbus.error import ErrorMapper
from dasbus.server.handler import ServerObjectHandler

import gi
gi.require_version("Gio", "2.0")
from gi.repository import Gio

log = logging.getLogger(__name__)

__all__ = [
    "GLibConnection",
    "MessageBus",
    "SystemMessageBus",
    "SessionMessageBus",
    "AddressedMessageBus"
]


class GLibConnection(object):
    """The low-level DBus connection library based on GLib."""

    DEFAULT_FLAGS = (
        Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT |
        Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION
    )

    @staticmethod
    def get_system_bus_connection(cancellable=None):
        """Get a system bus connection."""
        log.info("Connecting to the system bus.")
        return Gio.bus_get_sync(
            Gio.BusType.SYSTEM,
            cancellable
        )

    @staticmethod
    def get_session_bus_connection(cancellable=None):
        """Get a session bus connection."""
        log.info("Connecting to the session bus.")
        return Gio.bus_get_sync(
            Gio.BusType.SESSION,
            cancellable
        )

    @staticmethod
    def get_addressed_bus_connection(bus_address, flags=DEFAULT_FLAGS,
                                     observer=None, cancellable=None):
        """Get a connection to a bus at the specified address."""
        return Gio.DBusConnection.new_for_address_sync(
            bus_address,
            flags,
            observer,
            cancellable
        )


class AbstractMessageBus(metaclass=ABCMeta):
    """Abstract representation of a message bus.

    The property connection represents a connection to the bus. You can
    register a service name with register_service, or publish an object
    with publish_object and get a proxy of a remote object with get_proxy.
    """

    @property
    @abstractmethod
    def connection(self):
        """The DBus connection."""
        return None

    def check_connection(self):
        """Check if the connection is set up.

        :return: True if the connection is set up otherwise False
        """
        try:
            return self.connection is not None
        except Exception as e:  # pylint: disable=broad-except
            log.warning("Connection can't be created:\n%s", e)
            return False

    @abstractmethod
    def get_proxy(self, service_name, object_path, interface_name=None,
                  **kwargs):
        """Returns a proxy of a remote DBus object.

        :param service_name: a DBus name of a service
        :param object_path: a DBus path of an object
        :param interface_name: a DBus name of an interface or None
        :return: a proxy object
        """
        pass

    @abstractmethod
    def register_service(self, service_name, **kwargs):
        """Register a service on DBus.

        A service can be registered by requesting its name on DBus.
        This method should be called only after all of the required
        objects of the service are published on DBus.

        :param service_name: a DBus name of a service
        """
        pass

    @abstractmethod
    def publish_object(self, object_path, obj, **kwargs):
        """Publish an object on DBus.

        :param object_path: a DBus path of an object
        :param obj: an instance of @dbus_interface or @dbus_class
        """
        pass

    @abstractmethod
    def disconnect(self):
        """Disconnect from DBus."""
        pass


class MessageBus(AbstractMessageBus):
    """Representation of a message bus based on D-Bus."""

    def __init__(self, error_mapper=None, provider=GLibConnection):
        """Create a new message bus.

        :param error_mapper: a DBus error mapper
        :param provider: a provider of DBus connections
        """
        super().__init__()
        self._provider = provider
        self._error_mapper = error_mapper or ErrorMapper()
        self._connection = None
        self._proxy = None
        self._registrations = {}
        self._requested_names = set()

    @property
    def connection(self):
        """The DBus connection."""
        if not self._connection:
            self._connection = self._get_connection()

        return self._connection

    @abstractmethod
    def _get_connection(self):
        """Return a DBus connection."""
        pass

    @property
    def proxy(self):
        """The proxy of DBus."""
        if not self._proxy:
            self._proxy = self.get_proxy(
                "org.freedesktop.DBus",
                "/org/freedesktop/DBus"
            )

        return self._proxy

    # pylint: disable=arguments-differ
    def get_proxy(self, service_name, object_path, interface_name=None,
                  proxy_factory=None, **proxy_arguments):
        """Returns a proxy of a remote DBus object.

        If the proxy factory is not specified, we will use a default
        one. If the interface name is set, we will choose InterfaceProxy,
        otherwise ObjectProxy.

        If the interface name is set, we will add it to the additional
        arguments for the proxy factory.

        :param service_name: a DBus name of a service
        :param object_path: a DBus path of an object
        :param interface_name: a DBus name of an interface or None
        :param proxy_factory: a factory of a DBus object proxy
        :param proxy_arguments: additional arguments for the proxy factory
        :return: a proxy object
        """
        self._check_service_access(service_name)

        if not proxy_factory:
            proxy_factory = InterfaceProxy if interface_name else ObjectProxy

        if interface_name:
            proxy_arguments["interface_name"] = interface_name

        return proxy_factory(
            self,
            service_name,
            object_path,
            error_mapper=self._error_mapper,
            **proxy_arguments
        )

    def _check_service_access(self, service_name):
        """Check if we can access a DBus service.

        FIXME: This is a temporary check that should be later removed.

        This is useful during the transition of the Anaconda code from
        UI to DBus modules. This check prevents a deadlock in case that
        a DBus module tries to access a service, that it provides, from
        the main thread.

        :param service_name: a DBus name of a service
        :raises: RuntimeError if the service cannot be accessed
        """
        if service_name not in self._requested_names:
            # We don't provide this service.
            return

        if threading.current_thread() is not threading.main_thread():
            # We don't try to access this service from the main thread.
            return

        raise RuntimeError(
            "Can't access DBus service '{}' from "
            "the main thread.".format(service_name)
        )

    # pylint: disable=arguments-differ
    def register_service(self, service_name,
                         flags=DBUS_NAME_FLAG_ALLOW_REPLACEMENT):
        """Register a service on DBus.

        :param service_name: a DBus name of a service
        :param flags: the flags argument of the RequestName DBus method
        """
        log.debug("Registering a service name %s.", service_name)
        result = self.proxy.RequestName(service_name, flags)

        if result != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER:
            raise ConnectionError("Name request has failed: {}".format(result))

        self._requested_names.add(
            service_name
        )
        self._registrations[service_name] = (
            lambda: self.proxy.ReleaseName(service_name)
        )

    # pylint: disable=arguments-differ
    def publish_object(self, object_path, obj,
                       server_factory=ServerObjectHandler,
                       **server_arguments):
        """Publish an object on DBus.

        :param object_path: a DBus path of an object
        :param obj: an instance of @dbus_interface or @dbus_class
        :param server_factory: a factory of a DBus server object handler
        :param server_arguments: additional arguments for the server factory
        """
        log.debug("Publishing an object at %s.", object_path)
        object_handler = server_factory(
            self,
            object_path,
            obj,
            error_mapper=self._error_mapper,
            **server_arguments
        )
        object_handler.connect_object()

        self._registrations[object_path] = object_handler.disconnect_object

    def _unregister(self, object_path):
        """Remove Object from DBus."""

        if object_path in self._registrations:
            callback = self._registrations.pop(object_path)
            callback()

    def unpublish_object(self, object_path):
        log.debug("Removing object %s from the bus.", object_path)
        self._unregister(object_path)

    def unregister_service(self, object_path):
        log.debug("Removing service %s from the bus.", object_path)
        self._unregister(object_path)

    def disconnect(self):
        """Disconnect from DBus."""
        log.debug("Disconnecting from the bus.")

        while self._registrations:
            _, callback = self._registrations.popitem()
            callback()

        self._connection = None
        self._requested_names = set()


class SystemMessageBus(MessageBus):
    """Representation of a system bus connection."""

    def _get_connection(self):
        """Get a system DBus connection."""
        log.info("Connecting to the system bus.")
        return self._provider.get_system_bus_connection()


class SessionMessageBus(MessageBus):
    """Representation of a session bus connection."""

    def _get_connection(self):
        """Get a session DBus connection."""
        log.info("Connecting to the session bus.")
        return self._provider.get_session_bus_connection()


class AddressedMessageBus(MessageBus):
    """Representation of a connection for the specified address."""

    def __init__(self, address, *args, **kwargs):
        """Create a new representation of a connection.

        :param address: a bus address
        """
        super().__init__(*args, **kwargs)
        self._address = address

    @property
    def address(self):
        """The bus address."""
        return self._address

    def _get_connection(self):
        """Get a connection to a bus at the specified address."""
        log.info("Connecting to a bus at %s.", self._address)
        return self._provider.get_addressed_bus_connection(self._address)



#
# Support for Unix file descriptors.
#
# Copyright (C) 2022  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import logging

from dasbus.constants import DBUS_FLAG_NONE
from dasbus.typing import VariantUnpacking, get_variant
from dasbus.client.handler import GLibClient
from dasbus.server.handler import GLibServer

import gi
gi.require_version("Gio", "2.0")
from gi.repository import Gio

log = logging.getLogger(__name__)

__all__ = [
    "GLibClientUnix",
    "GLibServerUnix",
]


def acquire_fds(variant):
    """Acquire Unix file descriptors contained in a variant.

    Return a variant with indexes into a list of Unix file descriptors
    and the list of Unix file descriptors.

    If the variant is None, or the variant doesn't contain any Unix
    file descriptors, return None instead of the list.

    :param variant: a variant with Unix file descriptors
    :return: a variant with indexes and a list of Unix file descriptors
    """
    if variant is None:
        return None, None

    fd_list = []

    def _get_idx(fd):
        fd_list.append(fd)
        return len(fd_list) - 1

    variant_without_fds = UnixFDSwap.apply(variant, _get_idx)

    if not fd_list:
        return variant, None

    return variant_without_fds, Gio.UnixFDList.new_from_array(fd_list)


def restore_fds(variant, fd_list: Gio.UnixFDList):
    """Restore Unix file descriptors in a variant.

    If the variant is None, return None. Otherwise, return
    a variant with Unix file descriptors.

    :param variant: a variant with indexes into fd_list
    :param fd_list: a list of Unix file descriptors
    :return: a variant with Unix file descriptors
    """
    if variant is None:
        return None

    if fd_list is None:
        return variant

    fd_list = fd_list.steal_fds()

    if not fd_list:
        return variant

    def _get_fd(index):
        try:
            return fd_list[index]
        except IndexError:
            return -1

    return UnixFDSwap.apply(variant, _get_fd)


class UnixFDSwap(VariantUnpacking):
    """Class for swapping values of the UnixFD type."""

    @classmethod
    def apply(cls, variant, swap):
        """Swap unix file descriptors with indices.

        The provided function should swap a unix file
        descriptor with an index into an array of unix
        file descriptors or vice versa.

        :param variant: a variant to modify
        :param swap: a swapping function
        :return: a modified variant
        """
        return cls._recreate_variant(variant, swap)

    @classmethod
    def _handle_variant(cls, variant, *extras):
        """Handle a variant."""
        return cls._recreate_variant(variant.get_variant(), *extras)

    @classmethod
    def _handle_value(cls, variant, *extras):
        """Handle a basic value."""
        type_string = variant.get_type_string()

        # Handle the unix file descriptor.
        if type_string == 'h':
            # Get the swapping function.
            swap, *_ = extras
            # Swap the values.
            return swap(variant.get_handle())

        return variant.unpack()

    @classmethod
    def _recreate_variant(cls, variant, *extras):
        """Create a variant with swapped values."""
        type_string = variant.get_type_string()

        # Do nothing if there is no unix file descriptor to handle.
        if 'h' not in type_string and 'v' not in type_string:
            return variant

        # Get a new value of the variant.
        value = cls._process_variant(variant, *extras)

        # Create a new variant.
        return get_variant(type_string, value)


class GLibClientUnix(GLibClient):
    """The low-level DBus client library based on GLib."""

    @classmethod
    def sync_call(cls, connection, service_name, object_path, interface_name,
                  method_name, parameters, reply_type, flags=DBUS_FLAG_NONE,
                  timeout=GLibClient.DBUS_TIMEOUT_NONE):
        """Synchronously call a DBus method.

        :return: a result of the DBus call
        """
        # Process Unix file descriptors in parameters.
        parameters, fd_list = acquire_fds(parameters)

        # Call the DBus method.
        result = connection.call_with_unix_fd_list_sync(
            service_name,
            object_path,
            interface_name,
            method_name,
            parameters,
            reply_type,
            flags,
            timeout,
            fd_list,
            None
        )

        # Restore Unix file descriptors in the result.
        return restore_fds(*result)

    @classmethod
    def async_call(cls, connection, service_name, object_path, interface_name,
                   method_name, parameters, reply_type, callback,
                   callback_args=(), flags=DBUS_FLAG_NONE,
                   timeout=GLibClient.DBUS_TIMEOUT_NONE):
        """Asynchronously call a DBus method."""
        # Process Unix file descriptors in parameters.
        parameters, fd_list = acquire_fds(parameters)

        # Call the DBus method.
        connection.call_with_unix_fd_list(
            service_name,
            object_path,
            interface_name,
            method_name,
            parameters,
            reply_type,
            flags,
            timeout,
            fd_list,
            callback=cls._async_call_finish,
            user_data=(callback, callback_args)
        )

    @classmethod
    def _async_call_finish(cls, source_object, result_object, user_data):
        """Finish an asynchronous DBus method call."""
        # Prepare the user's callback.
        callback, callback_args = user_data

        def _finish_call():
            # Retrieve the result of the call.
            result = source_object.call_with_unix_fd_list_finish(
                result_object
            )
            # Restore Unix file descriptors in the result.
            return restore_fds(*result)

        # Call user's callback.
        callback(_finish_call, *callback_args)


class GLibServerUnix(GLibServer):
    """The low-level DBus server library based on GLib.

    Adds Unix FD Support to base class"""

    @classmethod
    def emit_signal(cls, connection, object_path, interface_name,
                    signal_name, parameters, destination=None):
        """Emit a DBus signal.

        GLib doesn't seem to support Unix file descriptors in signals.
        Swap Unix file descriptors with indexes into a list of Unix file
        descriptors, but emit just the indexes. Log a warning to inform
        users about the limited support.
        """
        # Process Unix file descriptors in parameters.
        parameters, fd_list = acquire_fds(parameters)

        if fd_list:
            log.warning("Unix file descriptors in signals are unsupported.")

        # Emit the signal without Unix file descriptors.
        connection.emit_signal(
            destination,
            object_path,
            interface_name,
            signal_name,
            parameters
        )

    @classmethod
    def set_call_reply(cls, invocation, out_type, out_value):
        """Set the reply of the DBus call."""
        # Process Unix file descriptors in the reply.
        reply_value = cls._get_reply_value(out_type, out_value)
        reply_args = acquire_fds(reply_value)

        # Send the reply.
        invocation.return_value_with_unix_fd_list(*reply_args)

    @classmethod
    def _object_callback(cls, connection, sender, object_path,
                         interface_name, method_name, parameters,
                         invocation, user_data):
        """A method call closure of a DBus object."""
        # Prepare the user's callback.
        callback, callback_args = user_data

        # Restore Unix file descriptors in parameters.
        fd_list = invocation.get_message().get_unix_fd_list()
        parameters = restore_fds(parameters, fd_list)

        # Call user's callback.
        callback(
            invocation,
            interface_name,
            method_name,
            parameters,
            *callback_args
        )



#
# Templates for DBus interfaces
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from abc import ABCMeta

from dasbus.server.property import PropertiesInterface

__all__ = [
    "BasicInterfaceTemplate",
    "InterfaceTemplate"
]


class BasicInterfaceTemplate(metaclass=ABCMeta):
    """Basic template for a DBus interface.

    This template uses a software design pattern called proxy.

    This class provides a recommended way how to define DBus interfaces
    and create publishable DBus objects. The class that defines a DBus
    interface should inherit this class and be decorated with @dbus_class
    or @dbus_interface decorator. The implementation of this interface will
    be provided by a separate object called implementation. Therefore the
    methods of this class should call the methods of the implementation,
    the signals should be connected to the signals of the implementation
    and the getters and setters of properties should access the properties
    of the implementation.

    .. code-block:: python

        @dbus_interface("org.myproject.X")
        class InterfaceX(BasicInterfaceTemplate):
            def DoSomething(self) -> Str:
                return self.implementation.do_something()

        class X(object):
            def do_something(self):
                return "Done!"

        x = X()
        i = InterfaceX(x)

        DBus.publish_object("/org/myproject/X", i)

    """

    def __init__(self, implementation):
        """Create a publishable DBus object.

        :param implementation: an implementation of this interface
        """
        self._implementation = implementation
        self.connect_signals()

    @property
    def implementation(self):
        """Return the implementation of this interface.

        :return: an implementation
        """
        return self._implementation

    def connect_signals(self):
        """Interconnect the signals.

        You should connect the emit methods of the interface
        signals to the signals of the implementation. Every
        time the implementation emits a signal, this interface
        reemits the signal on DBus.
        """
        pass


class InterfaceTemplate(BasicInterfaceTemplate, PropertiesInterface):
    """Template for a DBus interface.

    The interface provides the support for the standard interface
    org.freedesktop.DBus.Properties.

    Usage:

    .. code-block:: python

        def connect_signals(self):
            super().connect_signals()
            self.implementation.module_properties_changed.connect(
                self.flush_changes
            )
            self.watch_property("X", self.implementation.x_changed)

        @property
        def X(self, x) -> Int:
            return self.implementation.x

        @emits_properties_changed
        def SetX(self, x: Int):
            self.implementation.set_x(x)

    """

    def __init__(self, implementation):
        PropertiesInterface.__init__(self)
        BasicInterfaceTemplate.__init__(self, implementation)

    def watch_property(self, property_name, signal):
        """Watch a DBus property.

        Report a change when the property is changed.

        :param property_name: a name of a DBus property
        :param signal: a signal that emits when the property is changed
        """
        self._properties_changes.check_property(property_name)

        def callback(*args, **kwargs):
            self.report_changed_property(property_name)

        signal.connect(callback)



#
# Server support for publishable Python objects
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from abc import ABCMeta, abstractmethod

__all__ = ["Publishable"]


class Publishable(metaclass=ABCMeta):
    """Abstract class for Python objects that can be published on DBus.

    Example:

    .. code-block:: python

        # Define a publishable class.
        class MyObject(Publishable):

            def for_publication(self):
                return MyDBusInterface(self)

        # Create a publishable object.
        my_object = MyObject()

        # Publish the object on DBus.
        DBus.publish_object("/org/project/x", my_object.for_publication())

   """

    @abstractmethod
    def for_publication(self):
        """Return a DBus representation of this object.

        :return: an instance of @dbus_interface or @dbus_class
        """
        return None
